/*
    Copyright (C) 2002  bjk <bjk@arbornet.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>
#include <syslog.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <pwd.h>
#include <ctype.h>
#include "bubblegum.h"

int
isinteger (const char *str)
{
  int i;

  for (i = 0; i < strlen (str); i++)
    {
      if (isdigit (str[i]) == 0)
	return 1;
    }

  return 0;
}

void
logit (const char *format, ...)
{
  FILE *logfp;
  va_list ap;
  char *buf, timebuf[MAX_TIME_LEN];
  time_t now;
  struct tm *mytm;

  if (loglevel == 0 && usesyslog == 0)
    return;

  memset (timebuf, '\0', sizeof (timebuf));
  va_start (ap, format);

  if (usesyslog)
    {
      openlog (PROGNAME, LOG_PID, SYSLOG_FACILITY);
      vsyslog (SYSLOG_PRIORITY, format, ap);
      closelog ();

      va_end (ap);
      return;
    }

  if (!loglevel)
    {
      va_end (ap);
      return;
    }

  if ((logfp = fopen (logfile, "a")) == NULL)
    {
      va_end (ap);
      perror (logfile);
      return;
    }

  vasprintf (&buf, format, ap);
  va_end (ap);

  time (&now);
  mytm = localtime (&now);
  strftime (timebuf, sizeof (timebuf), TIMEFORMAT, mytm);

  fprintf (logfp, "%s %i %s\n", timebuf, pid, buf);

  fclose (logfp);
  free (buf);

  return;
}

void *
Realloc (void *ptr, size_t size)
{
  void *ptr2;

  if ((ptr2 = realloc (ptr, size)) == NULL)
    {
      logit ("%s(%i) malloc(): %s", __FILE__, __LINE__, strerror (errno));
      perror ("malloc()");
      exit (errno);
    }

  return ptr2;
}

void *
Malloc (size_t size)
{
  void *ptr;

  if ((ptr = malloc (size)) == NULL)
    {
      logit ("%s(%i) malloc(): %s", __FILE__, __LINE__, strerror (errno));
      perror ("malloc()");
      exit (errno);
    }

  return ptr;
}

void *
Calloc (size_t number, size_t size)
{
  void *ptr;

  if ((ptr = calloc (number, size)) == NULL)
    {
      logit ("%s(%i) calloc(): %s", __FILE__, __LINE__, strerror (errno));
      perror ("calloc()");
      exit (errno);
    }

  return ptr;
}

int
pathlength (const char *wd, const char *filename)
{
  if (filename[0] == '/')
    return strlen (filename);

  return (strlen (wd) + strlen (filename) + 1);
}

char *
parseargv (char **cmd)
{
  int len = 0, noquote = 0;
  char *str, *s;

  if ((s = *cmd) == NULL)
    return NULL;

  if (*s == 0)
    return NULL;

  str = Calloc (1, strlen (*cmd) + 1);

  while (*s)
    {
      int c;

      if (*s == ' ' && *s++)
	{
	  *cmd = s;
	  str -= len;
	  return str;
	}

      if (*s == '\\' && *++s == '\"')
	noquote = 1;

      if (*s == '\"' && !noquote)
	{
	  c = *s++;

	  do
	    {
	      if (*s == '\\' && *++s == '\"');
	      *str++ = *s++;
	      len++;
	    }
	  while (*s != '\"');

	  if (*++s == 0)
	    {
	      *cmd = NULL;
	      str -= len;
	      return str;
	    }

	  c = *s++;
	  *cmd = s;

	  str -= len;
	  return str;
	}

      *str++ = *s++;
      len++;
      noquote = 0;
    }

  *cmd = NULL;
  str -= len;
  return str;
}

char *
changetype (unsigned bits)
{
  char *str = NULL;
  static char buf[255];
  int bit;

  buf[0] = 0;

  for (bit = 1; bit < MAXTYPEBITS; bit++)
    {
      if (TEST_BIT (bits, bit))
	{
	  switch (bit)
	    {
	    case IDEV:
	      str = IDEV_STR;
	      break;
	    case INODE:
	      str = INODE_STR;
	      break;
	    case PERMS:
	      str = PERMS_STR;
	      break;
	    case UID:
	      str = UID_STR;
	      break;
	    case GID:
	      str = GID_STR;
	      break;
	    case SIZE:
	      str = SIZE_STR;
	      break;
	    case HLINKS:
	      str = HLINKS_STR;
	      break;
	    default:
	      str = NULL;
	      break;
	    }

	  if (str)
	    {
	      strncat (buf, str, sizeof (buf));
	      strncat (buf, ",", sizeof (buf));
	    }
	}
    }

  if (buf[0] == 0)
    return NULL;

  buf[strlen (buf) - 1] = 0;
  return buf;
}

int
is_file (const char *filename)
{
  struct stat mystat;

  if (stat (filename, &mystat) == -1)
    return -1;

  if (!(mystat.st_mode & S_IFREG))
    return 1;

  return 0;
}

char *
difftimestr (time_t compare, time_t basetime)
{
  int diff = 0;
  int day, hour, min, sec, i;
  static char buf[32];

  day = hour = min = sec = i = 0;
  diff = compare - basetime;

  day = diff / 86400;
  i = diff % 86400;

  hour = i / 3600;
  i = i % 3600;

  min = i / 60;
  i = i % 60;

  sec = i;

  snprintf (buf, sizeof (buf), "%i:%.2i:%.2i:%.2i", day, hour, min, sec);
  return buf;
}

char *
fullpath (const char *wd, char *filename, char dest[], size_t size)
{
  dest[0] = 0;

  if (filename[0] == '/')
    strncpy (dest, filename, size);
  else
    {
      strncpy (dest, wd, size);
      strncat (dest, "/", size);
      strncat (dest, filename, size);
    }

  return dest;
}

static char *
itoa (int i)
{
  static char buf[32];

  snprintf (buf, sizeof (buf), "%i", i);
  return strdup (buf);
}

char *
changelist (unsigned changes)
{
  static char buf[255];
  int bit;

  buf[0] = 0;

  for (bit = 1; bit < MAXCHGBITS; bit++)
    {
      if (TEST_BIT (changes, bit))
	{
	  switch (bit)
	    {
	    case ACCESS:
	      strncat (buf, ATIME_STR, sizeof (buf));
	      break;
	    case MODIFY:
	      strncat (buf, MTIME_STR, sizeof (buf));
	      break;
	    case CHANGE:
	      strncat (buf, CTIME_STR, sizeof (buf));
	      break;
	    case ERROR:
	      strncat (buf, ERROR_STR, sizeof (buf));
	      break;
	    default:
	      break;
	    }

	  strncat (buf, ",", sizeof (buf));
	}
    }

  buf[strlen (buf) - 1] = 0;
  return buf;
}

char *
parsecmd (struct filelist *flist, char *line, time_t basetime,
	  char *diff, unsigned changes, unsigned runs, unsigned runtotal)
{
  static char buf[sizeof (command)];
  int i = 0;

  buf[0] = 0;

  if (line[0] == 0)
    return NULL;

  while (*line)
    {
      struct tm *mytm;
      char *s = NULL;
      char timebuf[MAX_TIME_LEN] = { 0 };
      int n, malloced = 0, bufsize;
      time_t blah = 0;

      if (*line == '%')
	{
	  switch (*++line)
	    {
	    case 'i':
	      s = diff;
	      break;
	    case 'r':
	      bufsize = snprintf (NULL, 0, "%i/%i", runs, runtotal) + 1;
	      s = Calloc (1, bufsize);
	      malloced = 1;

	      snprintf (s, bufsize, "%i/%i", runs, runtotal);
	      break;
	    case 'f':
	      s = (char *) flist->filename;
	      break;
	    case 'F':
	      if (TEST_BIT (changes, ERROR))
		{
		  s = NONE_STR;
		  break;
		}

	      if (S_ISDIR (flist->laststat.st_mode))
		s = DIR_STR;
	      else if (S_ISCHR (flist->laststat.st_mode))
		s = CHR_STR;
	      else if (S_ISBLK (flist->laststat.st_mode))
		s = BLK_STR;
	      else if (S_ISREG (flist->laststat.st_mode))
		s = REG_STR;
	      else if (S_ISFIFO (flist->laststat.st_mode))
		s = FIFO_STR;
	      else if (S_ISLNK (flist->laststat.st_mode))
		s = LNK_STR;
	      else if (S_ISFIFO (flist->laststat.st_mode))
		s = SOCK_STR;
	      else
		s = "UNKNOWN";
	      break;
	    case 'm':
	      if (TEST_BIT (changes, ERROR))
		{
		  s = NONE_STR;
		  break;
		}

	      s = Calloc (1, 5);
	      malloced = 1;

	      snprintf (s, 5, "%.4o", flist->laststat.st_mode & ALLPERMS);
	      break;
	    case 't':
	      /*
	       * kinda hokey. use the latest change time. 
	       */
	      if (flist->laststat.st_atime > blah)
		blah = flist->laststat.st_atime;

	      if (flist->laststat.st_mtime > blah)
		blah = flist->laststat.st_mtime;

	      if (flist->laststat.st_ctime > blah)
		blah = flist->laststat.st_ctime;

	      mytm = localtime (&blah);

	      strftime (timebuf, sizeof (timebuf), TIMEFORMAT, mytm);
	      timebuf[strlen (timebuf)] = '\0';
	      s = (char *) timebuf;
	      break;
	    case 's':
	      bufsize = snprintf (NULL, 0,
				  "niceness: %i, loglvl: %i, int: %lis, "
				  "files: %i", niceness, loglevel,
				  interval, total) + 1;
	      s = Calloc (1, bufsize);
	      malloced = 1;

	      snprintf (s, bufsize,
			"niceness: %i, loglvl: %i, int: %lis, "
			"files: %i", niceness, loglevel, interval, total);
	      break;
	    case 'd':
	      time (&blah);
	      mytm = localtime (&blah);
	      strftime (timebuf, sizeof (timebuf), TIMEFORMAT, mytm);
	      timebuf[strlen (timebuf)] = '\0';

	      s = (char *) timebuf;
	      break;
	    case 'b':
	      if (TEST_BIT (changes, ERROR))
		{
		  s = NONE_STR;
		  break;
		}

	      mytm = localtime (&basetime);
	      strftime (timebuf, sizeof (timebuf), TIMEFORMAT, mytm);
	      timebuf[strlen (timebuf)] = '\0';

	      s = (char *) timebuf;
	      break;
	    case 'c':
	      s = changelist (changes);
	      break;
	    case 'y':
	      s = changetype (flist->chgtype);

	      if (!s)
		s = NONE_STR;
	      break;
	    case 'p':
	      s = itoa (pid);
	      malloced = 1;
	      break;
	    case 'e':
	      s = (flist->staterrno) ? strerror (flist->staterrno) : NONE_STR;
	      break;
	    case '%':
	      s = "%";
	      break;
	    default:
	      break;
	    }
	}

      if (s)
	{
	  for (n = 0; n < strlen (s); n++)
	    {
	      if (strlen (buf) + 1 == sizeof (buf))
		{
		  buf[sizeof (buf)] = 0;
		  return buf;
		}

	      buf[i++] = s[n];
	    }

	  if (malloced)
	    free (s);

	  s = NULL;
	  n = *line++;
	  continue;
	}

      if (strlen (buf) + 1 == sizeof (buf))
	break;

      buf[i++] = *line++;
    }

  buf[i] = 0;
  return buf;
}

void
freelist (struct filelist *flisthead)
{
  struct filelist *cur, *next;

  for (cur = flisthead; cur; cur = next)
    {
      next = cur->next;
      free (cur->filename);
      free (cur);
    }

  return;
}

void
catchsig (int signal)
{
  int status;

  switch (signal)
    {
    case SIGALRM:
      logit ("signal %i: timer expired, forcing check", signal);
      break;
    case SIGTERM:
      logit ("signal %i: terminating", signal);
      runlevel = QUIT;
      break;
    case SIGCHLD:
      waitpid (-1, &status, WNOHANG);

      /*
       * not reliable 
       */
      /*
       * if (WIFEXITED(status)) logit("command finished (%i): exit
       * status %i", cpid, WEXITSTATUS(status)); 
       */
      break;
    case SIGHUP:
      if (filenamelist)
	{
	  logit ("signal %i: reloading filelist", signal);
	  runlevel = RELOAD;
	}
      else
	logit ("signal %i: no file specified (-f)", signal);

      break;
    case SIGUSR1:
      logit ("signal %i: resetting all changed bits", signal);
      runlevel = RESET;
      break;
    case SIGUSR2:
      logit ("signal %i: resetting all changes bits and command runs",
	     signal);
      runlevel = RESETALL;
      break;
    default:
      logit ("signal %i: ignoring", signal);
      break;
    }

  return;
}

/* test to see if any bits are set */
int
checkbits (unsigned bits)
{
  int bit;

  for (bit = 1; bit < MAXCHGBITS; bit++)
    {
      if (TEST_BIT (bits, bit))
	return 1;
    }

  return 0;
}

char *
changelog (struct filelist *flist, time_t basetime, unsigned changes)
{
  char *line = NULL, *type = NULL, *timediff;
  int bit, error = 0, bufsize;
  time_t diff = 0;
  unsigned origchanged = flist->changed;

  bufsize = LINE_MAX;
  line = Calloc (1, bufsize);

  for (bit = 1; bit < MAXCHGBITS; bit++)
    {
      if (TEST_BIT (changes, bit) && TEST_BIT (check, bit))
	{
	  switch (bit)
	    {
	    case ACCESS:
	      strncat (line, ATIME_STR, bufsize);
	      strncat (line, ",", bufsize);

	      if (flist->laststat.st_atime > diff)
		diff = flist->laststat.st_atime;
	      break;
	    case MODIFY:
	      strncat (line, MTIME_STR, bufsize);
	      strncat (line, ",", bufsize);

	      if (flist->laststat.st_mtime > diff)
		diff = flist->laststat.st_mtime;
	      break;
	    case CHANGE:
	      strncat (line, CTIME_STR, bufsize);
	      strncat (line, ",", bufsize);
	      type = changetype (flist->chgtype);

	      if (flist->laststat.st_ctime > diff)
		diff = flist->laststat.st_ctime;
	      break;
	    case ERROR:
	      strncat (line, ERROR_STR, bufsize);
	      strncat (line, ",", bufsize);
	      error = 1;
	      break;
	    }

	  SET_BIT (flist->changed, bit);
	}
    }

  line[strlen (line) - 1] = '\0';

  if (type)
    {
      strncat (line, " ", bufsize);
      strncat (line, "(", bufsize);
      strncat (line, type, bufsize);
      strncat (line, ")", bufsize);
    }

  timediff = difftimestr (diff, basetime);

  if (loglevel == 1)
    {
      if (checkbits (origchanged))
	return timediff;
    }

  logit ("%s (%s) %s", line,
	 (error) ? strerror (flist->staterrno) : timediff, flist->filename);

  free (line);
  return (TEST_BIT (changes, ERROR)) ? NONE_STR : timediff;
}

void
resetbittypes (struct filelist *flist)
{
  RESET_BIT (flist->chgtype, IDEV);
  RESET_BIT (flist->chgtype, SIZE);
  RESET_BIT (flist->chgtype, INODE);
  RESET_BIT (flist->chgtype, HLINKS);
  RESET_BIT (flist->chgtype, UID);
  RESET_BIT (flist->chgtype, PERMS);
  RESET_BIT (flist->chgtype, GID);

  return;
}

void
resetchanged (struct filelist *flist)
{
  RESET_BIT (flist->changed, ACCESS);
  RESET_BIT (flist->changed, MODIFY);
  RESET_BIT (flist->changed, CHANGE);
  RESET_BIT (flist->changed, ERROR);

  return;
}

unsigned int
checkstat (struct filelist *flist, struct stat mystat,
	   time_t * basetime, unsigned changes)
{
  unsigned change = changes;

  if (TEST_BIT (check, ACCESS) && mystat.st_atime != flist->laststat.st_atime)
    {
      SET_BIT (change, ACCESS);
      *basetime = flist->laststat.st_atime;
    }

  if (TEST_BIT (check, MODIFY) && mystat.st_mtime != flist->laststat.st_mtime)
    {
      SET_BIT (change, MODIFY);
      *basetime = flist->laststat.st_mtime;
    }

  if (TEST_BIT (check, CHANGE) && mystat.st_ctime != flist->laststat.st_ctime)
    {
      SET_BIT (change, CHANGE);
      *basetime = flist->laststat.st_ctime;

      if (mystat.st_dev != flist->laststat.st_dev)
	SET_BIT (flist->chgtype, IDEV);

      if (mystat.st_ino != flist->laststat.st_ino)
	SET_BIT (flist->chgtype, INODE);

      if (mystat.st_nlink != flist->laststat.st_nlink)
	SET_BIT (flist->chgtype, HLINKS);

      if (mystat.st_uid != flist->laststat.st_uid)
	SET_BIT (flist->chgtype, UID);

      if (mystat.st_gid != flist->laststat.st_gid)
	SET_BIT (flist->chgtype, GID);

      if (mystat.st_size != flist->laststat.st_size)
	SET_BIT (flist->chgtype, SIZE);

      if (mystat.st_mode != flist->laststat.st_mode)
	SET_BIT (flist->chgtype, PERMS);
    }

  return change;
}

void
daemonstatus (unsigned cmdruns, unsigned runs, unsigned totalruns)
{
  char *tr = itoa (totalruns);
  char *r = itoa (runs);
  char *cr = itoa (cmdruns);

  logit ("lvl: %i, int: %is, files: %i, fileruns: %s, totalruns: %s/%s",
	 loglevel, interval, total,
	 (command[0] && !totalruns) ? cr : "-",
	 (command[0]) ? r : "-", (command[0] && totalruns) ? tr : "-");

  free (tr);
  free (r);
  free (cr);

  return;
}

void
doit (struct filelist *flisthead, unsigned cmdruns, unsigned totalruns)
{
  unsigned runs = 0;

  logit ("%s %s watching %s%s%s%s", PROGNAME, VERSION,
	 (TEST_BIT (check, ACCESS)) ? ATIME_STR " " : "",
	 (TEST_BIT (check, MODIFY)) ? MTIME_STR " " : "",
	 (TEST_BIT (check, CHANGE)) ? CTIME_STR " " : "",
	 (TEST_BIT (check, ERROR)) ? ERROR_STR : "");

  daemonstatus (cmdruns, runs, totalruns);

  while (1)
    {
      struct timeval tv;
      struct filelist *flist, *cur;
      int status;

      switch (runlevel)
	{
	case RELOAD:
	  if ((cur = loadfile (filenamelist, flisthead)) != NULL)
	    {
	      freelist (flisthead);
	      flisthead = cur;
	      daemonstatus (cmdruns, runs, totalruns);
	      break;
	    }

	  break;
	case RESETALL:
	case RESET:
	  for (cur = flisthead; cur; cur = flist)
	    {
	      flist = cur->next;
	      resetchanged (cur);

	      if (runlevel == RESETALL)
		cur->cmdruns = 0;
	    }

	  if (runlevel == RESETALL && totalruns)
	    runs = 0;

	  daemonstatus (cmdruns, runs, totalruns);
	  break;
	case QUIT:
	  freelist (flisthead);

	  if (filenamelist)
	    free (filenamelist);

	  free (logfile);
	  free (wd);
	  exit (EXIT_SUCCESS);
	  break;
	default:
	  break;
	}

      flist = flisthead;
      runlevel = NORMAL;

      while (flist->next != NULL)
	{
	  char **cargv;
	  char *cmd = NULL, *diff, *cp;
	  unsigned changes = 0;
	  int i = 0;
	  time_t basetime = 0;
	  struct stat mystat;

	  if (STAT (flist->filename, &mystat) == -1)
	    {
	      if (flist->staterrno == errno)
		{
		  flist = flist->next;
		  continue;
		}

	      if (TEST_BIT (check, ERROR))
		{
		  SET_BIT (changes, ERROR);
		  flist->staterrno = errno;
		}
	    }
	  else
	    flist->staterrno = 0;

	  changes = checkstat (flist, mystat, &basetime, changes);

	  if (checkbits (changes))
	    {
	      flist->laststat = mystat;
	      diff = changelog (flist, basetime, changes);

	      if (((totalruns && runs < totalruns) ||
		   (!cmdruns && !totalruns) ||
		   (flist->cmdruns < cmdruns && !totalruns)) && command[0])
		{
		  flist->cmdruns++;
		  runs++;

		  cmd = parsecmd (flist, command, basetime, diff,
				  changes,
				  (totalruns) ? runs : flist->cmdruns,
				  (totalruns) ? totalruns : cmdruns);

		  switch ((cpid = fork ()))
		    {
		    case 0:
		      i = 0;
		      cargv = Calloc (1, sizeof (char *));

		      while ((cp = parseargv (&cmd)) != NULL)
			{
			  cargv = Realloc (cargv, (i + 2) * sizeof (char *));
			  cargv[i++] = strdup (cp);
			  free (cp);
			}

		      cargv[i] = NULL;
		      errno = EXIT_SUCCESS;

		      if (execvp (cargv[0], cargv) == -1)
			{
			  logit ("%s: %s", cargv[0], strerror (errno));
			  _exit (errno);
			}

		      _exit (EXIT_FAILURE);
		    case -1:
		      logit ("child: fork(): %s", strerror (errno));
		      _exit (errno);
		    default:
		      logit ("%i/%i (pid %i): %s",
			     (totalruns) ? runs : flist->cmdruns,
			     (totalruns) ? totalruns : cmdruns, cpid, cmd);

		      waitpid (-1, &status, WNOHANG);
		      break;
		    }
		}

	      resetbittypes (flist);

	      if (loglevel == 2)
		resetchanged (flist);
	    }

	  flist = flist->next;
	}

      tv.tv_sec = interval;
      tv.tv_usec = 0;

      select (0, 0, 0, 0, &tv);
    }

  return;
}

static void
usage (const char *pn)
{
  fprintf (stderr, "Usage: %s [-hv] [-k] [-s] [-l <logfile>] [-level] "
	   "[-e <command>]\n\t[-r runs | -t runs] [-n niceness] "
	   "-A | -amcE -f <filelist> | file ...\n", pn);
  fprintf (stderr, "  -a\tCheck the access time of the file.\n");
  fprintf (stderr, "  -m\tCheck the modification time of the file.\n");
  fprintf (stderr, "  -c\tCheck the change time of the inode.\n");
  fprintf (stderr, "  -E\tCheck for errors when stat()ing a file.\n");
  fprintf (stderr, "  -A\tCheck everything (shortcut for -amcE).\n");
  fprintf (stderr, "  -e\tCommand to run when a file changes (see below).\n");
  fprintf (stderr,
	   "  -r\tNumber of times to run command per file "
	   "(0=infinite). The default is %i.\n", CMDRUNS);
  fprintf (stderr,
	   "  -t\tTotal number of times to run command (overrides " "-r).\n");
  fprintf (stderr,
	   "  -n\tNiceness level (-20 through 20) of the daemon and "
	   "command.\n");
  fprintf (stderr,
	   "  -i\tInterval in seconds to check files. The default is "
	   "%i.\n", INTERVAL);
  fprintf (stderr,
	   "  -f\tA file containing a list of filenames, one per line, "
	   "to watch.\n");
  fprintf (stderr, "  -k\tDon't follow symbolic links.\n");
  fprintf (stderr, "  -l\tLog to an alternate file. The default is "
	   "~/%s.\n", LOGFILE);
  fprintf (stderr,
	   "  -#\tLog level: 0=none, 1=first change, 2=all changes. "
	   "Default is %i.\n", LOGLEVEL);
  fprintf (stderr,
	   "  -s\tDisable logging to file and log to " "syslogd(8).\n");
  fprintf (stderr, "  -h\tThis help text.\n");
  fprintf (stderr, "  -v\tVersion and copyright information.\n\n");
  fprintf (stderr, "The following expansions are available in a command "
	   "(remember to quote!):\n");
  fprintf (stderr, "  %%f - filename\t");
  fprintf (stderr, "\t%%t - time of (latest) change\t");
  fprintf (stderr, "%%b - last change time\n");
  fprintf (stderr, "  %%e - error info\t");
  fprintf (stderr, "%%i - time difference\t\t");
  fprintf (stderr, "%%p - daemon PID\n");
  fprintf (stderr, "  %%d - detection time\t");
  fprintf (stderr, "%%s - daemon status\t\t");
  fprintf (stderr, "%%r - run number/total\n");
  fprintf (stderr, "  %%F - file type\t");
  fprintf (stderr, "%%m - file permission mode\t");
  fprintf (stderr, "%%%% - regular %%\n");
  fprintf (stderr, "  %%c - change(s): %s, %s, %s, %s\n", ATIME_STR,
	   MTIME_STR, CTIME_STR, ERROR_STR);
  fprintf (stderr, "  %%y - change type(s): %s, %s, %s, %s, %s, %s, %s\n",
	   IDEV_STR, INODE_STR, HLINKS_STR, UID_STR, GID_STR, PERMS_STR,
	   SIZE_STR);

  exit (EXIT_FAILURE);
}

int
main (int argc, char *argv[])
{
  FILE *fp;
  int opt, bufsize = 0;
  char *pn;
  unsigned cmdruns, totalruns;
  static struct sigaction sa;
  struct filelist *flisthead, *flist;
  struct passwd *passwd;

  if ((wd = getcwd (NULL, 0)) == NULL)
    {
      perror ("getcwd()");
      exit (errno);
    }

  if ((passwd = getpwuid (getuid ())) == NULL)
    {
      fprintf (stderr, "getpwuid(): unknown error\n");
      exit (EXIT_FAILURE);
    }

  bufsize = snprintf (NULL, 0, "%s/%s", passwd->pw_dir, LOGFILE) + 1;
  logfile = Malloc (bufsize);
  snprintf (logfile, bufsize, "%s/%s", passwd->pw_dir, LOGFILE);

  interval = INTERVAL;
  loglevel = LOGLEVEL;
  cmdruns = CMDRUNS;
  totalruns = 0;
  symlinks = 0;
  pn = PROGNAME;

  errno = 0;
  pid = getpid ();

  if ((niceness = getpriority (PRIO_PROCESS, pid)) == -1)
    {
      if (errno)
	{
	  perror ("getpriority()");
	  exit (errno);
	}
    }

  while ((opt = getopt (argc, argv, "An:t:skhvr:aEe:i:mf:l:c012")) != -1)
    {
      int err;
      char tmp[2];

      switch (opt)
	{
	case 'k':
	  symlinks = 1;
	  break;
	case 'n':
	  niceness = atoi (optarg);

	  if (niceness > 20 || niceness < -20)
	    usage (pn);
	  break;
	case 's':
	  usesyslog = 1;
	  break;
	case 'e':
	  if (optarg[0] == 0)
	    usage (pn);

	  strncpy (command, optarg, sizeof (command));
	  break;
	case 'l':
	  bufsize = pathlength (wd, optarg) + 1;
	  logfile = Realloc (logfile, bufsize);
	  fullpath (wd, optarg, logfile, bufsize);
	  break;
	case '0':
	case '1':
	case '2':
	  snprintf (tmp, sizeof (tmp), "%c", opt);
	  loglevel = atoi (tmp);
	  break;
	case 'f':
	  bufsize = pathlength (wd, optarg) + 1;
	  filenamelist = Malloc (bufsize);
	  fullpath (wd, optarg, filenamelist, bufsize);

	  if ((err = is_file (filenamelist)) != 0)
	    {
	      if (err < 0)
		{
		  perror (filenamelist);
		  exit (errno);
		}
	      else
		fprintf (stderr, "%s: not a file\n", filenamelist);

	      exit (EXIT_FAILURE);
	    }
	  break;
	case 't':
	  if (isinteger (optarg))
	    usage (pn);

	  totalruns = atoi (optarg);
	  break;
	case 'r':
	  if (isinteger (optarg))
	    usage (pn);

	  cmdruns = atoi (optarg);
	  break;
	case 'i':
	  if (isinteger (optarg))
	    usage (pn);

	  if ((interval = atol (optarg)) < 1)
	    usage (pn);
	  break;
	case 'h':
	  usage (pn);
	  break;
	case 'v':
	  fprintf (stdout, "%s %s\n%s\n", PROGNAME, VERSION, COPYRIGHT);
	  exit (EXIT_SUCCESS);
	case 'A':
	  SET_BIT (check, ACCESS);
	  SET_BIT (check, CHANGE);
	  SET_BIT (check, MODIFY);
	  SET_BIT (check, ERROR);
	  break;
	case 'a':
	  SET_BIT (check, ACCESS);
	  break;
	case 'c':
	  SET_BIT (check, CHANGE);
	  break;
	case 'm':
	  SET_BIT (check, MODIFY);
	  break;
	case 'E':
	  SET_BIT (check, ERROR);
	  break;
	default:
	  usage (pn);
	  break;
	}
    }

  if ((argc == optind && !filenamelist) || !check)
    usage (pn);

  if (setpriority (PRIO_PROCESS, pid, niceness) == -1)
    {
      perror ("setpriority()");
      exit (errno);
    }

  if (!command[0] && !usesyslog && !loglevel)
    fprintf (stdout, "No logging OR command? Whatever you say...\n");

  if (!usesyslog && loglevel)
    {
      if ((fp = fopen (logfile, "a")) == NULL)
	{
	  perror (logfile);
	  exit (errno);
	}

      fclose (fp);
    }

  if (filenamelist)
    {
      if ((flisthead = loadfile (filenamelist, NULL)) == NULL)
	{
	  perror (filenamelist);
	  exit (errno);
	}
    }
  else
    {
      flisthead = Calloc (1, sizeof (struct filelist));
      flist = flisthead;

      while (optind < argc)
	{
	  struct stat mystat;
	  char *tmp, *cp;

	  if (STAT (argv[optind], &mystat) == -1)
	    {
	      perror (argv[optind++]);
	      continue;
	    }

	  bufsize = pathlength (wd, argv[optind]) + 1;
	  tmp = Malloc (bufsize);
	  cp = fullpath (wd, argv[optind++], tmp, bufsize);
	  flist->filename = Calloc (1, bufsize);

	  strncpy (flist->filename, cp, bufsize);
	  flist->laststat = mystat;

	  flist->next = Calloc (1, sizeof (struct filelist));
	  flist = flist->next;

	  total++;
	  free (tmp);
	}

      flist->next = NULL;
    }

  if (!flisthead->filename)
    {
      fprintf (stderr, "No files to watch. Exiting.\n");
      exit (EXIT_FAILURE);
    }

  sigfillset (&sa.sa_mask);
  sigdelset (&sa.sa_mask, SIGTERM);
  sigdelset (&sa.sa_mask, SIGCHLD);
  sigdelset (&sa.sa_mask, SIGUSR1);
  sigdelset (&sa.sa_mask, SIGUSR2);
  sigdelset (&sa.sa_mask, SIGHUP);
  sigdelset (&sa.sa_mask, SIGALRM);
  sigdelset (&sa.sa_mask, SIGILL);
  sigdelset (&sa.sa_mask, SIGBUS);
  sigdelset (&sa.sa_mask, SIGFPE);

  if (sigprocmask (SIG_BLOCK, &sa.sa_mask, 0) == -1)
    {
      perror ("sigprocmask()");
      exit (errno);
    }

  sa.sa_handler = catchsig;
#ifdef SA_NOCLDWAIT
  sa.sa_flags = SA_NOCLDWAIT;
#else
  sa.sa_flags = 0;
#endif

  if (sigaction (SIGHUP, &sa, NULL) == -1)
    {
      perror ("sigaction()");
      exit (errno);
    }

  if (sigaction (SIGALRM, &sa, NULL) == -1)
    {
      perror ("sigaction()");
      exit (errno);
    }

  if (sigaction (SIGTERM, &sa, NULL) == -1)
    {
      perror ("sigaction()");
      exit (errno);
    }

  if (sigaction (SIGCHLD, &sa, NULL) == -1)
    {
      perror ("sigaction()");
      exit (errno);
    }

  if (sigaction (SIGUSR1, &sa, NULL) == -1)
    {
      perror ("sigaction()");
      exit (errno);
    }

  if (sigaction (SIGUSR2, &sa, NULL) == -1)
    {
      perror ("sigaction()");
      exit (errno);
    }

  switch (fork ())
    {
    case 0:
      if (chdir ("/") == -1)
	{
	  perror ("/");
	  exit (errno);
	}

      if (setsid () == -1)
	{
	  perror ("setsid()");
	  exit (errno);
	}

      freopen ("/dev/null", "r", stdin);
      freopen ("/dev/null", "w", stdout);
      freopen ("/dev/null", "w", stderr);

      doit (flisthead, cmdruns, totalruns);
      break;
    case -1:
      perror ("fork()");
      exit (errno);
      break;
    default:
      break;
    }

  exit (EXIT_SUCCESS);
}
